# Copyright (C) 2021-2022 Hiroshi Shinaoka and others
# SPDX-License-Identifier: MIT
import numpy as np
from typing import Dict, Optional, Union, Tuple, List, cast, Sequence, Any

from sparse_ir import MatsubaraSampling, TauSampling

from admmsolver.objectivefunc import LeastSquares, ConstrainedLeastSquares, \
    ObjectiveFunctionBase, L2Regularizer, \
    SemiPositiveDefinitePenalty, L1Regularizer
from admmsolver.optimizer import SimpleOptimizer, Model, EqualityCondition
from admmsolver.matrix import identity, PartialDiagonalMatrix, \
    ScaledIdentityMatrix, MatrixBase, DenseMatrix, DiagonalMatrix

from enum import Enum


class InputType(Enum):
    TIME = 0
    FREQ = 1


class AnaContBase:
    r"""
    Base solver for analytic continuation

    We model this analytic-continuation problem as
        G_{s,ij} = - \sum_l S_l U_{sl} * rho_{l,ij}
                        + alpha^p * |\sum_l B_{tl} * rho_{l,ij}|_p^p
                        + \sum_n C_{sn} x'_{n,ij},
    where
        s: an imaginary time or an imaginary frequency,
        i and j: spin orbitals,
        alpha: a regularization,
        S_l: singular values.
    The first and last terms of the RHS denote a normal component
    and a singular (augment) component, respectively.

    The expansion coefficients in IR are given by
        rho_{l,ij} = \sum_m A_{l,m} x_{m,ij}.
    The sum rules reads
        \sum_l S_{k,l} rho_{l,ij} = T_{l,ij}.

    The singular component represents an additonal contribution
    that is not complactly (or conveniently) represented in IR
    (e.g., Hartree-Fock term, a delta peak at omega=0 for bosons).
    """
    def __init__(
            self,
            sampling: Sequence[Union[TauSampling, MatsubaraSampling]],
            a: MatrixBase,
            b: MatrixBase,
            c: MatrixBase,
            reg_type: str = "L2",
            sum_rule: Optional[Tuple[np.ndarray, np.ndarray]] = None
            ) -> None:
        r"""
        sampling:
            A sequence of sampling objects.
            The first/second component is regarded as
            a normal/singular component.

        a:
            Transformation matrix `A` from the normal component `x` to rho_l

        b:
            Transformation matrix `B` from the normal component `x`
            to the space where L1/L2 regularation is imposed.

        c:
            Transformation matrix `C` from the normal component `x`
            to the space where the SPD condition is imposed.
            This matrix must be consistent with `smpl_reqal_w`.

        reg_type: str
            `L1` or `L2`

        sum_rule: (S, T)
            Sum rule for the normal component.
            S, T are 2D and 1D array, respectively
        """
        assert a.shape[1] == b.shape[1], f"{a.shape} {b.shape}"
        assert a.shape[1] == c.shape[1], f"{a.shape} {c.shape}"
        assert reg_type in ["L1", "L2"]
        assert len(sampling) == 1 or len(sampling) == 2
        assert np.unique([s_.sampling_points.size for s_ in sampling]).size \
            == 1

        self._sampling = sampling

        self._bases = [s.basis for s in sampling]
        self._beta = cast(float, self._bases[0])  # type: float
        self._is_augmented = len(sampling) > 1

        self._a = a
        self._b = b
        self._c = c
        self._reg_type = reg_type
        self._sum_rule = sum_rule

    def rho_omega(
            self,
            x: np.ndarray,
            omega: np.ndarray) -> np.ndarray:
        """ Compute rho(omega) """
        assert x.ndim == 3
        assert omega.ndim == 1
        rho_l = np.einsum(
            'lx,xij->lij',
            self._a.asmatrix(),
            x[0:self._a.shape[1], ...],
            optimize=True)
        v_omega = self._bases[0].v(omega)
        return np.einsum('lw,lij->wij', v_omega, rho_l, optimize=True)

    def solve_elbow(
            self,
            ginput: np.ndarray,
            alpha_min: float,
            alpha_max: float,
            n_alpha: int,
            **kwargs) -> Tuple[np.ndarray, Dict]:
        assert alpha_min < alpha_max
        assert n_alpha > 0
        alphas = np.exp(
            np.linspace(np.log(alpha_max), np.log(alpha_min), n_alpha))

        lstsq_alpha = []
        info = {}  # type: Dict[str, Any]
        info["info"] = {}
        info["x"] = {}
        initial_guess = None  # type: Optional[List[np.ndarray]]
        for alpha in alphas:
            x_, info_ = self.solve(
                ginput, alpha, initial_guess=initial_guess, **kwargs)
            lstsq_alpha.append(info_["lstsq"])
            info["x"][alpha] = x_
            info["info"][alpha] = info_
            initial_guess = info_["optimizer"].x
        info["alphas"] = alphas
        info["lstsq_alphas"] = np.asarray(lstsq_alpha)

        # Find optimal alpha
        x = np.log(alphas)
        y = np.log(lstsq_alpha)
        a = (y[-1]-y[0])/(x[-1]-x[0])
        y = -(y - (a*(x-x[0]) + y[0]))
        ialpha_opt = np.argmax(y)
        info["ialpha_opt"] = ialpha_opt

        return info["x"][alphas[ialpha_opt]], info

    def solve(
            self,
            ginput: np.ndarray,
            alpha: float,
            niter: int = 10000,
            spd: bool = True,
            interval_update_mu: int = 100,
            rtol: float = 1e-10,
            initial_guess: Sequence[np.ndarray] = None):
        r"""
        ginput: 3d array of shape (nsmpl, nf, nf),
            where nv is the number of Matsubara frequeicies/times,
            and nf is the number of flavors.

        input_type:
            'tau' or 'matsubara'

        alpha: float
            L2 regularization parameter

        niter: int
            Max number of iterations

        spd:
            Use semi-positive-definite condition for normal component

        interval_update_mu:
            Interval for updating mu

        rtol:
            Stopping condition.
            Check if all relative norms of all primal and dual residuals
            are smaller than rtol.

        initial_guess:
            Initial guess
        """
        assert ginput.ndim == 3
        assert ginput.shape[0] == self._sampling[0].sampling_points.size, \
            "shape of ginput is not consistent with sampling frequencies!"
        assert ginput.shape[1] == ginput.shape[2], \
            "Invalid shape of ginput"
        if isinstance(self._sum_rule, np.ndarray):
            assert ginput.shape[1:] == self._sum_rule.shape,\
                f"{ginput.shape[1:]} {self._sum_rule.shape}"

        nf = ginput.shape[1]  # type: int
        nparam_normal = self._a.shape[1] * nf**2
        assert isinstance(nparam_normal, int)
        nparam_singular = 0 if not self._is_augmented else nf**2
        assert isinstance(nparam_singular, int)
        x_size = nparam_normal + nparam_singular
        assert isinstance(x_size, int)

        ###
        # Fitting matrix
        # Let (T0, T1) = self._sampling.matrix.a.
        #
        # Normal component:
        #  g_smpl = - (T0 @ S @ A0) x
        #
        # Singular component:
        #  g_smpl = T1 x'
        ###
        T0 = self._sampling[0].matrix.a
        T0S = T0.copy()
        T0S[:, 0:self._bases[0].size] *= -self._bases[0].s[None, :]
        fitmat_normal = T0S @ self._a.asmatrix()  # type: np.ndarray
        if self._is_augmented:
            fitmat_full = PartialDiagonalMatrix(
                    np.hstack((fitmat_normal, self._sampling[1].matrix.a)),
                    (nf, nf)
                )
        else:
            fitmat_full = PartialDiagonalMatrix(fitmat_normal, (nf, nf))

        ###
        # Various contraints V*x = W on the normal component
        ###
        V = None  # Optional[MatrixBase]
        W = None  # Optional[np.ndarray]

        # Sum-rule constraint
        if self._sum_rule is not None:
            a, b = self._sum_rule
            assert a.ndim == 2 and b.ndim == 2
            if self._is_augmented:
                a = cast(np.ndarray, _add_zero_column(a))
            V = PartialDiagonalMatrix(a, (nf, nf))
            W = b.ravel()

        equality_conditions = []  # type: List[EqualityCondition]

        ###
        #  Regularization
        ###
        reg = None  # type: Optional[ObjectiveFunctionBase]
        if self._reg_type == "L1":
            # x1 = b * x0 and |x1|_1
            b_ = self._b
            if self._is_augmented:
                b_ = _add_zero_column(b_)
            b_full = PartialDiagonalMatrix(b_, (nf, nf))
            equality_conditions.append(
                EqualityCondition(
                    0, 1, b_full, identity(int(b_full.shape[0]))
                ),
            )
            reg = L1Regularizer(alpha, int(b_full.shape[0]))
        elif self._reg_type == "L2":
            # x1 = x0 and |b * x1|_2^2
            if isinstance(self._b, DenseMatrix):
                b_arr_ = self._b.asmatrix()  # type: np.ndarray
                if self._is_augmented:
                    b_arr_ = cast(np.ndarray, _add_zero_column(b_arr_))
                b_full = PartialDiagonalMatrix(b_arr_, (nf, nf))
            elif isinstance(self._b, ScaledIdentityMatrix):
                b_ = self._b
                if self._is_augmented:
                    b_ = _add_zero_column(b_)
                b_full = PartialDiagonalMatrix(b_, (nf, nf))
            else:
                raise RuntimeError("Unsupported!")
            equality_conditions.append(
                EqualityCondition(
                    0, 1, identity(x_size), identity(x_size)
                ),
            )
            reg = L2Regularizer(alpha, b_full)
        assert reg is not None

        # Optimizer
        lstsq = None  # type: Optional[ObjectiveFunctionBase]
        if V is not None:
            assert V is not None and W is not None
            lstsq = ConstrainedLeastSquares(
                1.0, fitmat_full, ginput.ravel(),
                V, W
            )
        else:
            lstsq = LeastSquares(1.0, fitmat_full, ginput.ravel())

        terms = [lstsq, reg]  # type: List[ObjectiveFunctionBase]
        if spd:
            # Semi-positive definite condition on normal component
            nsmpl_w = self._c.shape[0]
            nn = SemiPositiveDefinitePenalty((nsmpl_w, nf, nf), 0)
            c_ext = _add_zero_column(self._c) \
                if self._is_augmented else self._c
            equality_conditions.append(
                    EqualityCondition(
                        0, len(terms),
                        PartialDiagonalMatrix(c_ext, (nf, nf)),
                        PartialDiagonalMatrix(
                            ScaledIdentityMatrix(nsmpl_w, 1.0), (nf, nf)),
                    )
                )
            terms.append(nn)

        model = Model(terms, equality_conditions)
        opt = SimpleOptimizer(model, x0=initial_guess)

        # Run
        opt.solve(niter, interval_update_mu=interval_update_mu, rtol=rtol)

        info = {
            "optimizer": opt,
            "lstsq": lstsq(opt.x[0])
        }

        x = opt.x[0].reshape((-1,) + ginput.shape[1:])
        return x, info


class SimpleAnaContBaseL2:
    r"""
    Analytic continuation with L2 regularization

    Augmentation is not supported

    We model this analytic-continuation problem as
        G_{s,ij} = - \sum_l S_l U_{sl} * rho_{l,ij}
                        + alpha * |\sum_l B_{tl} * rho_{l,ij}|_p^p
    where
        s: an imaginary time or an imaginary frequency,
        i and j: spin orbitals,
        alpha: a regularization,
        S_l: singular values.

    The expansion coefficients in IR are given by
        rho_{l,ij} = \sum_m A_{l,m} x_{m,ij}.
    The sum rules reads
        \sum_l S_{k,l} rho_{l,ij} = T_{l,ij}.
    """
    def __init__(
            self,
            sampling: Union[TauSampling, MatsubaraSampling],
            a: MatrixBase,
            b: MatrixBase,
            c: MatrixBase,
            sum_rule: Optional[Tuple[np.ndarray, np.ndarray]] = None,
            scale_alpha: bool = True
            ) -> None:
        r"""
        sampling:
            A sequence of sampling objects.
            The first/second component is regarded as
            a normal/singular component.

        a:
            Transformation matrix `A` from the normal component `x` to rho_l

        b:
            Transformation matrix `B` from the normal component `x`
            to the space where L1/L2 regularation is imposed.

        c:
            Transformation matrix `C` from the normal component `x`
            to the space where the SPD condition is imposed.
            This matrix must be consistent with `smpl_reqal_w`.

        reg_type: str
            `L1` or `L2`

        sum_rule: (S, T)
            Sum rule for the normal component.
            S, T are 2D and 1D array, respectively
        """
        assert a.shape[1] == b.shape[1], f"{a.shape} {b.shape}"
        assert a.shape[1] == c.shape[1], f"{a.shape} {c.shape}"

        self._sampling = sampling

        self._basis = sampling.basis
        self._beta = cast(float, self._basis)  # type: float

        self._a = a
        self._b = b
        self._c = c
        self._sum_rule = sum_rule


        ###
        # Fitting matrix
        # Let (T0, T1) = self._sampling.matrix.a.
        #
        # Normal component:
        #  g_smpl = - (T0 @ S @ A0) x
        #
        # Singular component:
        #  g_smpl = T1 x'
        ###
        T0 = self._sampling.matrix.a
        T0S = T0.copy()
        T0S[:, 0:self._basis.size] *= -self._basis.s[None, :]
        self._fitmat = T0S @ a.asmatrix()  # type: np.ndarray
        assert isinstance(b, DenseMatrix)

        # alpha
        self._coeff_alpha = 1.0
        if scale_alpha:
            self._coeff_alpha = cast(float, np.linalg.norm(self._fitmat, ord="fro")/
                np.linalg.norm(b.asmatrix(), ord="fro"))

    def rho_omega(
            self,
            x: np.ndarray,
            omega: np.ndarray) -> np.ndarray:
        """ Compute rho(omega) """
        assert x.ndim == 3
        assert omega.ndim == 1
        rho_l = np.einsum(
            'lx,xij->lij',
            self._a.asmatrix(),
            x[0:self._a.shape[1], ...],
            optimize=True)
        v_omega = self._basis.v(omega)
        return np.einsum('lw,lij->wij', v_omega, rho_l, optimize=True)

    def solve_elbow(
            self,
            ginput: np.ndarray,
            alpha_min: float,
            alpha_max: float,
            n_alpha: int,
            **kwargs) -> Tuple[np.ndarray, Dict]:
        assert alpha_min < alpha_max
        assert n_alpha > 0
        alphas = np.exp(
            np.linspace(np.log(alpha_max), np.log(alpha_min), n_alpha))

        lstsq_alpha = []
        info = {}  # type: Dict[str, Any]
        info["info"] = {}
        info["x"] = {}
        initial_guess = None  # type: Optional[List[np.ndarray]]
        for alpha in alphas:
            x_, info_ = self.solve(
                ginput, alpha, initial_guess=initial_guess, **kwargs)
            lstsq_alpha.append(info_["lstsq"])
            info["x"][alpha] = x_
            info["info"][alpha] = info_
            initial_guess = info_["optimizer"].x
        info["alphas"] = alphas
        info["lstsq_alphas"] = np.asarray(lstsq_alpha)

        # Find optimal alpha
        x = np.log(alphas)
        y = np.log(lstsq_alpha)
        a = (y[-1]-y[0])/(x[-1]-x[0])
        y = -(y - (a*(x-x[0]) + y[0]))
        ialpha_opt = np.argmax(y)
        info["ialpha_opt"] = ialpha_opt

        return info["x"][alphas[ialpha_opt]], info

    def solve(
            self,
            ginput: np.ndarray,
            alpha: float,
            niter: int = 10000,
            spd: bool = True,
            interval_update_mu: int = 100,
            rtol: float = 1e-10,
            initial_guess: Sequence[np.ndarray] = None):
        r"""
        ginput: 3d array of shape (nsmpl, nf, nf),
            where nv is the number of Matsubara frequeicies/times,
            and nf is the number of flavors.

        input_type:
            'tau' or 'matsubara'

        alpha: float
            L2 regularization parameter

        niter: int
            Max number of iterations

        spd:
            Use semi-positive-definite condition for normal component

        interval_update_mu:
            Interval for updating mu

        rtol:
            Stopping condition.
            Check if all relative norms of all primal and dual residuals
            are smaller than rtol.

        initial_guess:
            Initial guess
        """
        assert ginput.ndim == 3
        assert ginput.shape[0] == self._sampling.sampling_points.size, \
            "shape of ginput is not consistent with sampling frequencies!"
        assert ginput.shape[1] == ginput.shape[2], \
            "Invalid shape of ginput"
        if isinstance(self._sum_rule, np.ndarray):
            assert ginput.shape[1:] == self._sum_rule.shape,\
                f"{ginput.shape[1:]} {self._sum_rule.shape}"

        nf = ginput.shape[1]  # type: int
        nparam_normal = self._a.shape[1] * nf**2
        assert isinstance(nparam_normal, int)
        x_size = nparam_normal
        assert isinstance(x_size, int)

        ###
        # Fitting matrix
        # Let (T0, T1) = self._sampling.matrix.a.
        #
        # Normal component:
        #  g_smpl = - (T0 @ S @ A0) x
        #
        # Singular component:
        #  g_smpl = T1 x'
        ###
        #T0 = self._sampling.matrix.a
        #T0S = T0.copy()
        #T0S[:, 0:self._basis.size] *= -self._basis.s[None, :]
        #fitmat = T0S @ self._a.asmatrix()  # type: np.ndarray
        #assert isinstance(self._b, DenseMatrix)
        fitmat_ext = np.vstack((self._fitmat, (self._coeff_alpha*alpha) * self._b.asmatrix()))
        fitmat_full = PartialDiagonalMatrix(fitmat_ext, (nf, nf))

        fitvec_full = np.vstack([ginput, np.zeros((self._b.shape[0], nf, nf))])

        ###
        # Various contraints V*x = W on the normal component
        ###
        V = None  # Optional[MatrixBase]
        W = None  # Optional[np.ndarray]

        # Sum-rule constraint
        if self._sum_rule is not None:
            a, b = self._sum_rule
            assert a.ndim == 2 and b.ndim == 2
            V = PartialDiagonalMatrix(a, (nf, nf))
            W = b.ravel()

        equality_conditions = []  # type: List[EqualityCondition]

        # Optimizer
        lstsq = None  # type: Optional[ObjectiveFunctionBase]
        if V is not None:
            assert V is not None and W is not None
            lstsq = ConstrainedLeastSquares(
                1.0, fitmat_full, fitvec_full.ravel(),
                V, W
            )
        else:
            lstsq = LeastSquares(1.0, fitmat_full, fitvec_full.ravel())

        # If SPD is not used, a closed-form solution is available.
        if not spd:
            return lstsq.solve().reshape((-1,) + ginput.shape[1:]), {}

        terms = [lstsq]  # type: List[ObjectiveFunctionBase]
        # Semi-positive definite condition on normal component
        nsmpl_w = self._c.shape[0]
        nn = SemiPositiveDefinitePenalty((nsmpl_w, nf, nf), 0)
        equality_conditions.append(
                EqualityCondition(
                    0, len(terms),
                    PartialDiagonalMatrix(self._c, (nf, nf)),
                    PartialDiagonalMatrix(
                        ScaledIdentityMatrix(nsmpl_w, 1.0), (nf, nf)),
                )
            )
        terms.append(nn)

        model = Model(terms, equality_conditions)
        opt = SimpleOptimizer(model, x0=initial_guess)

        # Run
        opt.solve(niter, interval_update_mu=interval_update_mu, rtol=rtol)

        info = {
            "optimizer": opt,
            "lstsq": lstsq(opt.x[0])
        }

        x = opt.x[0].reshape((-1,) + ginput.shape[1:])
        return x, info


# a: (N, M)
# y: (N, 1)
# np.hstack((a,y)): (N, M+1)
def _add_zero_column(a: Union[np.ndarray, MatrixBase]):
    if isinstance(a, np.ndarray):
        assert a.ndim == 2
        N = a.shape[0]
        y = np.zeros((N, 1))
        return np.hstack((a, y))
    elif isinstance(a, ScaledIdentityMatrix):
        diagonals = np.full(min(*a.shape), a.coeff)
        return DiagonalMatrix(diagonals, shape=(a.shape[0], a.shape[1]+1))
    elif isinstance(a, DenseMatrix):
        return DenseMatrix(_add_zero_column(a.asmatrix()))
    else:
        raise RuntimeError(f"Invalid type{type(a)}!")


# Copyright (C) 2021-2022 H. Shinaoka and others
# SPDX-License-Identifier: MIT

__license__ = "MIT"
__version__ = "0.8"


from .smooth import AnaContSmooth, AnaContSmoothOpt
from .spm import AnaContSpM



# Copyright (C) 2021-2022 Hiroshi Shinaoka and others
# SPDX-License-Identifier: MIT
import numpy as np
from typing import Optional, Union, Tuple, Dict, List

from sparse_ir import FiniteTempBasis, MatsubaraSampling, TauSampling
from sparse_ir.augment import LegendreBasis, MatsubaraConstBasis

from admmsolver.matrix import DenseMatrix, ScaledIdentityMatrix

from .solver_base import AnaContBase, InputType
from .util import oversample


class AnaContSpM(object):
    r"""
    Analytic continuation solver based on SpM
    """
    def __init__(
            self,
            beta: float,
            wmax: float,
            statistics: str,
            input_type: Union[str, InputType],
            sampling_points: np.ndarray,
            moment: Optional[np.ndarray] = None,
            singular_term: Union[None, str] = None,
            eps=1e-15,
            oversampling=1,
            reg_type: str = "L1") -> None:
        r"""
        To be written...

        moment: None or 2d array
            First moment m_{ij} = int domega rho_{ij}(omega)

        """
        assert isinstance(beta, float)
        assert isinstance(wmax, float)
        assert isinstance(eps, float)
        assert type(input_type) in [InputType, str]
        if isinstance(input_type, str):
            input_type = {"time": InputType.TIME, "freq": InputType.FREQ}[input_type]
        assert isinstance(sampling_points, np.ndarray) and \
            sampling_points.ndim == 1
        assert moment is None or isinstance(moment, np.ndarray)

        basis = FiniteTempBasis(statistics, beta, wmax, eps=eps)

        # Fitting parameters (i.e, rho_l) to rho_l
        a = ScaledIdentityMatrix(basis.size, coeff=1.0)

        b = ScaledIdentityMatrix(basis.size, coeff=1.0)

        # From rho_l to the integral of rho(omega)
        sum_rule = None
        if moment is not None:
            assert moment.ndim == 2 and moment.shape[0] == moment.shape[1]
            sum_rule_coeff = basis.s * (basis.u(0) + basis.u(basis.beta))
            sum_rule_coeff = sum_rule_coeff.reshape((1,-1))
            sum_rule = (sum_rule_coeff, moment)

        bases = [basis] #type: List[Union[FiniteTempBasis, LegendreBasis, MatsubaraConstBasis]]
        if singular_term is not None:
            if singular_term == "omega0":
                assert statistics == "B"
                bases.append(LegendreBasis(statistics, beta, 1))
            elif singular_term == "HF":
                bases.append(MatsubaraConstBasis(statistics, beta, 1.0))
            else:
                raise RuntimeError(f"Invalid singular_term {singular_term}")

        sampling = []
        smpl_cls = {InputType.FREQ: MatsubaraSampling, InputType.TIME: TauSampling}[input_type]
        for b_ in bases:
            sampling.append(smpl_cls(b_, sampling_points))

        # For semi positive definite condition
        self._smpl_real_w = oversample(np.hstack((-wmax, basis.v[-1].roots(), wmax)), oversampling)
        assert bases[0].v is not None
        c = DenseMatrix(bases[0].v(self._smpl_real_w).T)

        self._solver = AnaContBase(
                sampling, a, b, c,
                sum_rule=sum_rule,
                reg_type=reg_type
            )

    @property
    def smpl_real_w(self):
        return self._smpl_real_w


    def rho_omega(
            self,
            x: np.ndarray,
            omega: np.ndarray
        ) -> np.ndarray:
        """ Compute rho(omega) """
        assert x.ndim == 3
        assert omega.ndim == 1
        rho_l = np.einsum(
            'lx,xij->lij',
            self._solver._a.asmatrix(),
            x[0:self._solver._a.shape[1],...],
            optimize=True)
        v_omega = self._solver._bases[0].v(omega)
        return np.einsum('lw,lij->wij', v_omega, rho_l, optimize=True)

    def solve(
            self,
            ginput: np.ndarray,
            alpha: float,
            niter: int = 10000,
            spd: bool = True,
            interval_update_mu: int =100,
            rtol: float = 1e-10
        ) -> Tuple[np.ndarray, Dict]:
        return self._solver.solve(ginput, alpha, niter, spd, interval_update_mu, rtol)


# Copyright (C) 2021-2022 Hiroshi Shinaoka and others
# SPDX-License-Identifier: MIT
import numpy as np
from typing import Optional, Union, Tuple, Dict, List
from scipy.interpolate import interp1d

from sparse_ir import FiniteTempBasis, MatsubaraSampling, TauSampling
from sparse_ir.augment import LegendreBasis, MatsubaraConstBasis

from admmsolver.matrix import DenseMatrix, ScaledIdentityMatrix
from admmsolver.util import smooth_regularizer_coeff

from .solver_base import AnaContBase, InputType, SimpleAnaContBaseL2
from .util import oversample, prj_w_to_l


class AnaContSmooth(object):
    r"""
    To be written...
    """
    def __init__(
            self,
            beta: float,
            wmax: float,
            statistics: str,
            input_type: Union[str, InputType],
            sampling_points: np.ndarray,
            oversampling: int = 1,
            moment: Optional[np.ndarray] = None,
            singular_term: Union[None, str] = None,
            reg_type: str = "L2",
            eps: float = 1e-15) -> None:
        r"""
        To be written...

        moment: None or 2d array
            First moment m_{ij} = int domega rho_{ij}(omega)

        See docstring of AnaContBase for other options
        """
        assert isinstance(beta, float)
        assert isinstance(wmax, float)
        assert isinstance(eps, float)
        assert type(input_type) in [InputType, str]
        if isinstance(input_type, str):
            input_type = {"time": InputType.TIME, "freq": InputType.FREQ}[input_type]
        assert isinstance(sampling_points, np.ndarray) and \
            sampling_points.ndim == 1
        assert isinstance(oversampling, int)
        assert moment is None or isinstance(moment, np.ndarray)

        basis = FiniteTempBasis(statistics, beta, wmax, eps=eps)

        # Fitting parameters for the normal component (sampling points in the real-frequency space)
        self._smpl_real_w = oversample(np.hstack((-wmax, basis.v[-1].roots(), wmax)), oversampling)

        # From rho(omega_i) to rho_l (using linear interpolation)
        a = prj_w_to_l(basis, self._smpl_real_w, 10)

        # From rho(omea_i) to rho''(omega) [Check uniform mesh]
        b = smooth_regularizer_coeff(self._smpl_real_w) # type: np.ndarray

        # From rho(omega_i) to the integral of rho(omega)
        sum_rule = None
        if moment is not None:
            assert moment.ndim == 2 and moment.shape[0] == moment.shape[1]
            sum_rule_coeff = basis.s * (basis.u(0) + basis.u(basis.beta))
            sum_rule_coeff = sum_rule_coeff.reshape((1,-1))
            sum_rule = (sum_rule_coeff, moment)

        bases = [basis] #type: List[Union[FiniteTempBasis, LegendreBasis, MatsubaraConstBasis]]
        if singular_term is not None:
            if singular_term == "omega0":
                assert statistics == "B"
                bases.append(LegendreBasis(statistics, beta, 1))
            elif singular_term == "HF":
                bases.append(MatsubaraConstBasis(statistics, beta, 1.0))
            else:
                raise RuntimeError(f"Invalid singular_term {singular_term}")

        sampling = []
        smpl_cls = {InputType.FREQ: MatsubaraSampling, InputType.TIME: TauSampling}[input_type]
        for b_ in bases:
            sampling.append(smpl_cls(b_, sampling_points))

        c = ScaledIdentityMatrix(self._smpl_real_w.size, 1.0)

        self._solver = AnaContBase(
                sampling,
                DenseMatrix(a),
                DenseMatrix(b),
                c,
                sum_rule=sum_rule,
                reg_type=reg_type
            )

    @property
    def smpl_real_w(self):
        return self._smpl_real_w

    def rho_omega(
            self,
            x: np.ndarray,
            omega: np.ndarray) -> np.ndarray:
        """ Compute rho(omega) """
        intp = interp1d(
            self._smpl_real_w, x[0:self._smpl_real_w.size, :, :], axis=0)
        return intp(omega)

    def solve(
            self,
            ginput: np.ndarray,
            alpha: float,
            niter: int = 10000,
            spd: bool = True,
            interval_update_mu: int = 100,
            rtol: float = 1e-10) -> Tuple[np.ndarray, Dict]:
        return self._solver.solve(
            ginput, alpha, niter, spd, interval_update_mu, rtol)

    def solve_elbow(
            self,
            ginput: np.ndarray,
            alpha_min: float,
            alpha_max: float,
            n_alpha: int,
            **kwargs) -> Tuple[np.ndarray, Dict]:
        return self._solver.solve_elbow(
            ginput, alpha_min, alpha_max, n_alpha, **kwargs)


class AnaContSmoothOpt(object):
    r"""
    To be written...
    """
    def __init__(
            self,
            beta: float,
            wmax: float,
            statistics: str,
            input_type: Union[str, InputType],
            sampling_points: np.ndarray,
            oversampling: int = 1,
            moment: Optional[np.ndarray] = None,
            eps: float = 1e-15,
            scale_alpha: bool = False) -> None:
        r"""
        To be written...

        moment: None or 2d array
            First moment m_{ij} = int domega rho_{ij}(omega)

        See docstring of AnaContBase for other options
        """
        assert isinstance(beta, float)
        assert isinstance(wmax, float)
        assert isinstance(eps, float)
        assert type(input_type) in [InputType, str]
        if isinstance(input_type, str):
            input_type = {
                "time": InputType.TIME, "freq": InputType.FREQ}[input_type]
        assert isinstance(sampling_points, np.ndarray) and \
            sampling_points.ndim == 1
        assert isinstance(oversampling, int)
        assert moment is None or isinstance(moment, np.ndarray)

        basis = FiniteTempBasis(statistics, beta, wmax, eps=eps)
        self.basis = basis

        # Fitting parameters for the normal component
        # (sampling points in the real-frequency space)
        self._smpl_real_w = oversample(
            np.hstack((-wmax, basis.v[-1].roots(), wmax)), oversampling)

        # From rho(omega_i) to rho_l (using linear interpolation)
        a = prj_w_to_l(basis, self._smpl_real_w, 10)

        # From rho(omea_i) to rho''(omega) [Check uniform mesh]
        b = smooth_regularizer_coeff(self._smpl_real_w)  # type: np.ndarray

        # From rho(omega_i) to the integral of rho(omega)
        sum_rule = None
        if moment is not None:
            assert moment.ndim == 2 and moment.shape[0] == moment.shape[1]
            sum_rule_coeff = basis.s * (basis.u(0) + basis.u(basis.beta))
            sum_rule_coeff = sum_rule_coeff.reshape((1, -1))
            sum_rule = (sum_rule_coeff, moment)

        sampling = []
        smpl_cls = {
            InputType.FREQ: MatsubaraSampling,
            InputType.TIME: TauSampling}[input_type]
        sampling = smpl_cls(basis, sampling_points)

        c = ScaledIdentityMatrix(self._smpl_real_w.size, 1.0)


        self._solver = SimpleAnaContBaseL2(
                sampling,
                DenseMatrix(a),
                DenseMatrix(b),
                c,
                sum_rule=sum_rule,
                scale_alpha=scale_alpha
            )
        self._a = a

    @property
    def smpl_real_w(self):
        return self._smpl_real_w

    def rho_l(self, x: np.ndarray) -> np.ndarray:
        """ Convert solution to expansion coefficients in V_l(omega) """
        assert x.ndim == 3
        assert x.shape[0] == self._a.shape[-1], f"{x.shape} {self._a.shape}"
        return np.einsum('Ww,wij->Wij', self._a, x)

    def g_l(self, x: np.ndarray) -> np.ndarray:
        """ Convert solution to expansion coefficients in U_l(omega) """
        return - self.basis.s[:, None, None] * self.rho_l(x)

    def rho_omega(
            self,
            x: np.ndarray,
            omega: np.ndarray) -> np.ndarray:
        """ Compute rho(omega) """
        intp = interp1d(
            self._smpl_real_w, x[0:self._smpl_real_w.size, :, :], axis=0)
        return intp(omega)

    def solve(
            self,
            ginput: np.ndarray,
            alpha: float,
            niter: int = 10000,
            spd: bool = True,
            interval_update_mu: int = 100,
            rtol: float = 1e-10) -> Tuple[np.ndarray, Dict]:
        return self._solver.solve(
            ginput, alpha, niter, spd, interval_update_mu, rtol)

    def solve_elbow(
            self,
            ginput: np.ndarray,
            alpha_min: float,
            alpha_max: float,
            n_alpha: int,
            **kwargs) -> Tuple[np.ndarray, Dict]:
        return self._solver.solve_elbow(
            ginput, alpha_min, alpha_max, n_alpha, **kwargs)



import numpy as np
from numpy.polynomial.legendre import leggauss
from spm_omega.quad import scale_quad, composite_leggauss

def test_scale_quad():
    """ Integrate x^3 over [0, 3] """
    xmax = 3
    x_, w_ = leggauss(10)
    x, w = scale_quad(x_, w_, 0, xmax)
    f = lambda x: x**3
    res = np.sum(w * f(x))
    ref = 0.25*(xmax**4)
    assert np.abs(res - ref) < 1e-10

def test_composite_leggauss():
    xmax = 3
    segments = np.array([0, 0.1, 3.0])
    deg = 10
    x, w = composite_leggauss(segments, deg)
    f = lambda x: x**3
    res = np.sum(w * f(x))
    ref = 0.25*(xmax**4)
    assert np.abs(res - ref) < 1e-10


from typing import Optional

from spm_omega import AnaContSmooth, AnaContSpM

import numpy as np
from sparse_ir import FiniteTempBasis,\
    MatsubaraSampling, TauSampling
import pytest
from itertools import product

from spm_omega.smooth import AnaContSmoothOpt


def gaussian(x, mu, sigma): return np.exp(-((x-mu)/sigma)**2) / \
    (np.sqrt(np.pi)*sigma)


def overlap(v, rho):
    r = rho(np.array([0.0, 0.0]))
    if isinstance(r, np.ndarray):
        assert r.ndim == 3
        shape = r.shape[1:]
        res = np.zeros((v.size, ) + shape, dtype=np.complex128)
        print("shape", shape)
        for i, j in product(range(shape[0]), range(shape[1])):
            #f = lambda x: rho(x)[i,j]
            #print("debug", rho(0.0))
            #print("debug", f(0.0))
            #print("debug", v.overlap(lambda x: rho(x)[0,i,j]).shape, v.size)
            res[:, i,j] = v.overlap(lambda x: rho(x)[0,i,j]).ravel()
        return res
    else:
        return v.overlap(rho)


def rho_single_orb(omega):
    res = 0.2*gaussian(omega, 0.0, 0.15) + \
        0.4*gaussian(omega, 1.0, 0.8) + 0.4*gaussian(omega, -1.0, 0.8)
    if isinstance(res, np.ndarray):
        return res[:, None, None]
    else:
        return res[None, None, None]


def rho_two_orb(omega):
    omega = np.asarray(omega)
    theta = 0.2*np.pi
    rot_mat = np.array([[np.cos(theta), -np.sin(theta)],
                       [np.sin(theta), np.cos(theta)]])
    rho_omega = np.zeros((omega.size, 2, 2))
    rho_omega[:, 0, 0] = 0.2*gaussian(omega, 0.0, 0.15)
    rho_omega[:, 1, 1] = 0.4 * \
        gaussian(omega, 1.0, 0.8) + 0.4*gaussian(omega, -1.0, 0.8)
    return np.einsum('ji,wjk,kl->wil', rot_mat, rho_omega, rot_mat)


def get_singular_term_matsubara(type: Optional[str], ginput: np.ndarray):
    if type == "HF":
        return np.ones_like(ginput)
    return np.zeros_like(ginput)


@pytest.mark.parametrize("stat", ["F", "B"])
@pytest.mark.parametrize("rho", [(rho_single_orb), (rho_two_orb)])
def test_smooth_solver(stat, rho):
    wmax = 10.0
    beta = 100.0
    eps = 1e-7
    alpha = 1e-8
    niv = 1000
    shift = {"F": 1, "B": 0}[stat]
    vsample = 2*np.arange(-niv, niv) + shift
    tausample = np.linspace(0, beta, 2*niv)
    niter = 1000

    basis = FiniteTempBasis(stat, beta, wmax, eps=eps)
    smpl_matsu = MatsubaraSampling(basis, vsample)
    smpl_tau = TauSampling(basis, tausample)

    # Compute exact rho_l, g_l, g_iv, g_tau
    rho_test = rho(np.linspace(-1, 1, 100))
    rho_l = overlap(basis.v, rho)
    rho_l = rho_l.reshape((basis.size,) + rho_test.shape[1:])
    g_l = -basis.s[:, None, None] * rho_l

    # From Matsubara
    g_iv = smpl_matsu.evaluate(g_l, axis=0)
    solver = AnaContSmoothOpt(beta, wmax, stat, "freq", vsample, eps=eps)
    x, info = solver.solve(g_iv, alpha, niter=niter, spd=True)
    rho_w = solver.rho_omega(x, solver.smpl_real_w)
    for x_, y_ in zip(
            rho_w[0:solver.smpl_real_w.size], rho(solver.smpl_real_w)):
        print(x_[0, 0].real, y_[0, 0].real)
    np.testing.assert_allclose(
        rho_w[0:solver.smpl_real_w.size],
        rho(solver.smpl_real_w), rtol=0, atol=0.15)

    # From tau
    gtau = smpl_tau.evaluate(g_l, axis=0)
    solver = AnaContSmoothOpt(beta, wmax, stat, "time", tausample, eps=eps)
    x, _ = solver.solve(gtau, alpha, niter=niter)
    rho_w = solver.rho_omega(x, solver.smpl_real_w)
    np.testing.assert_allclose(rho_w, rho(
        solver.smpl_real_w), rtol=0, atol=0.15)


def _test_solver(SolverType, stat, rho, augment, reg_type):
    wmax = 10.0
    beta = 100.0
    eps = 1e-7
    alpha = 1e-8
    niv = 1000
    shift = {"F": 1, "B": 0}[stat]
    vsample = 2*np.arange(-niv, niv) + shift
    tausample = np.linspace(0, beta, 2*niv)
    niter = {"L1": 1000, "L2": 1000}[reg_type]

    basis = FiniteTempBasis(stat, beta, wmax, eps=eps)
    smpl_matsu = MatsubaraSampling(basis, vsample)
    smpl_tau = TauSampling(basis, tausample)

    # Compute exact rho_l, g_l, g_iv, g_tau
    rho_test = rho(np.linspace(-1, 1, 100))
    rho_l = overlap(basis.v, rho)
    rho_l = rho_l.reshape((basis.size,) + rho_test.shape[1:])
    g_l = -basis.s[:, None, None] * rho_l

    # From Matsubara
    g_iv = smpl_matsu.evaluate(g_l, axis=0)
    g_iv += get_singular_term_matsubara(augment, g_iv)
    solver = SolverType(
        beta, wmax, stat, "freq", vsample, singular_term=augment,
        reg_type=reg_type, eps=eps
    )
    x, info = solver.solve(g_iv, alpha, niter=niter, spd=True)
    rho_w = solver.rho_omega(x, solver.smpl_real_w)
    for x_, y_ in zip(
            rho_w[0:solver.smpl_real_w.size], rho(solver.smpl_real_w)):
        print(x_[0, 0].real, y_[0, 0].real)
    np.testing.assert_allclose(
        rho_w[0:solver.smpl_real_w.size],
        rho(solver.smpl_real_w), rtol=0, atol=0.15)

    # From tau
    if augment != "HF":
        gtau = smpl_tau.evaluate(g_l, axis=0)
        solver = AnaContSmooth(beta, wmax, stat, "time",
                               tausample, reg_type=reg_type, eps=eps)
        x, _ = solver.solve(gtau, alpha, niter=niter)
        rho_w = solver.rho_omega(x, solver.smpl_real_w)
        np.testing.assert_allclose(rho_w, rho(
            solver.smpl_real_w), rtol=0, atol=0.15)


@pytest.mark.parametrize("stat", ["F", "B"])
@pytest.mark.parametrize("rho", [(rho_single_orb), (rho_two_orb)])
@pytest.mark.parametrize("augment", [None, "HF"])
@pytest.mark.parametrize("reg_type", ["L1", "L2"])
def test_smooth(stat, rho, augment, reg_type):
    _test_solver(AnaContSmooth, stat, rho, augment, reg_type)


@pytest.mark.parametrize("stat", ["F", "B"])
@pytest.mark.parametrize("rho", [(rho_single_orb), (rho_two_orb)])
@pytest.mark.parametrize("augment", [None, "HF"])
@pytest.mark.parametrize("reg_type", ["L1"])
def test_spm(stat, rho, augment, reg_type):
    _test_solver(AnaContSpM, stat, rho, augment, reg_type)


def test_elbow():
    stat = "F"
    wmax = 10.0
    beta = 100.0
    eps = 1e-7
    alpha_min = 1e-10
    alpha_max = 1.0
    n_alpha = 5
    niv = 500
    shift = {"F": 1, "B": 0}[stat]
    vsample = 2*np.arange(-niv, niv) + shift
    niter = 1000

    basis = FiniteTempBasis(stat, beta, wmax, eps=eps)
    smpl_matsu = MatsubaraSampling(basis, vsample)

    # Compute exact rho_l, g_l, g_iv, g_tau
    rho_test = rho_single_orb(np.linspace(-1, 1, 100))
    rho_l = overlap(basis.v, rho_single_orb)
    rho_l = rho_l.reshape((basis.size,) + rho_test.shape[1:])
    g_l = -basis.s[:, None, None] * rho_l

    # From Matsubara
    g_iv = smpl_matsu.evaluate(g_l, axis=0)
    solver = AnaContSmooth(beta, wmax, stat, "freq", vsample, eps=eps)
    x, info = solver.solve_elbow(
        g_iv, alpha_min, alpha_max, n_alpha, niter=niter)

#    for alpha, lst in zip(info["alphas"], info["lstsq_alphas"]):
#        print(alpha, lst)
    rho_w = solver.rho_omega(x, solver.smpl_real_w)
#    for x_, y_ in zip(
#            rho_w[0:solver.smpl_real_w.size],
#            rho_single_orb(solver.smpl_real_w)):
#        print(x_[0, 0].real, y_[0, 0].real)
    np.testing.assert_allclose(
        rho_w[0:solver.smpl_real_w.size],
        rho_single_orb(solver.smpl_real_w), rtol=0, atol=0.15)



import numpy as np
from numpy.polynomial.legendre import leggauss

def scale_quad(x, w, xmin, xmax):
    """ Scale weights and notes of quadrature to the interval [xmin, xmax] """
    assert xmin < xmax
    dx = xmax - xmin
    w_ = 0.5 * dx * w
    x_ = (0.5 * dx) * (x + 1) + xmin
    return x_, w_

def composite_leggauss(segments, deg):
    nsp = segments.size
    x, w = leggauss(deg)
    x_comp, w_comp = [], []
    for s in range(nsp-1):
        x_, w_ = scale_quad(x, w, segments[s], segments[s+1])
        x_comp.append(x_)
        w_comp.append(w_)
    return np.hstack(x_comp), np.hstack(w_comp)


import numpy as np
from numpy.polynomial.legendre import leggauss
from scipy.integrate import quad
from scipy.interpolate import interp1d

from sparse_ir import FiniteTempBasis
from .quad import scale_quad


def __oversample(x: np.ndarray):
    xmid = 0.5*(x[1:] + x[:-1])
    return np.unique(np.hstack((x, xmid)))


def oversample(x: np.ndarray, n: int = 1):
    for i in range(n):
        x = __oversample(x)
    return x


def prj_w_to_l(
        basis: FiniteTempBasis,
        smpl_w: np.ndarray,
        deg: int) -> np.ndarray:
    """ Projector from rho(omega_i) to rho_l """
    prj_w_to_l = np.zeros((basis.size, smpl_w.size))
    x_, w_ = leggauss(deg)
    for s in range(smpl_w.size-1):
        x, w = scale_quad(x_, w_, smpl_w[s], smpl_w[s+1])
        dx = smpl_w[s+1] - smpl_w[s]
        f = (x - smpl_w[s])/dx
        g = (smpl_w[s+1] - x)/dx
        for idx_l in range(basis.size):
            prj_w_to_l[idx_l, s+1] += np.sum(w * basis.v[idx_l](x) * f)
            prj_w_to_l[idx_l, s] += np.sum(w * basis.v[idx_l](x) * g)
    return prj_w_to_l



def regtarded_g(
        omegas_out: np.ndarray,
        rho_mesh,
        rho, epsabs=1e-5):
    """
    Compute retarded Green's function from spectral function

    omegas_out:
        real frequencies where G^R(w) is evaluated

    rho_mesh:
        real frequecies where rho(w) is given (in ascending order)

    rho:
        3D array representing rho(w).
        The first axis corresponds to the frequency.
    """
    assert rho.ndim == 3
    nso = rho.shape[1]
    res = np.zeros((omegas_out.size, nso, nso), dtype=np.complex128)

    interp = interp1d(rho_mesh, rho[:, 0, 0].real)

    for i in range(nso):
        for j in range(nso):
            interp = interp1d(
                        rho_mesh, rho[:, i, j].real,
                        fill_value=(0.0, 0.0), bounds_error=False)
            for k in range(omegas_out.size):
                res[k, i, j] += quad(
                                interp, rho_mesh[0], rho_mesh[-1],
                                weight="cauchy", wvar=omegas_out[k],
                                limit=1000, epsabs=epsabs)[0]
                res[k, i, j] += -1j*np.pi * interp(omegas_out[k])

            interp = interp1d(
                        rho_mesh, rho[:, i, j].imag,
                        fill_value=(0.0, 0.0), bounds_error=False)
            for k in range(omegas_out.size):
                res[k, i, j] += 1j*quad(
                                    interp,
                                    rho_mesh[0], rho_mesh[-1],
                                    weight="cauchy", wvar=omegas_out[k],
                                    limit=1000, epsabs=epsabs)[0]
                res[k, i, j] += np.pi * interp(omegas_out[k])
    return res



# Copyright (C) 2021-2022 Hiroshi Shinaoka and others
# SPDX-License-Identifier: MIT
import io, os.path, re
from setuptools import setup, find_packages


def readfile(*parts):
    """Return contents of file with path relative to script directory"""
    herepath = os.path.abspath(os.path.dirname(__file__))
    fullpath = os.path.join(herepath, *parts)
    with io.open(fullpath, 'r') as f:
        return f.read()


def extract_version(*parts):
    """Extract value of __version__ variable by parsing python script"""
    initfile = readfile(*parts)
    version_re = re.compile(r"(?m)^__version__\s*=\s*['\"]([^'\"]*)['\"]")
    match = version_re.search(initfile)
    assert match is not None
    return match.group(1)


VERSION = extract_version('src', 'spm_omega', '__init__.py')
REPO_URL = "https://github.com/shinaoka/spm_omega/"
LONG_DESCRIPTION = readfile('README.md')

setup(
    name='spm_omega',
    version=VERSION,

    description=
        'Sparse-modeling solvers for analytic continuation',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=' '.join([
        'sparse modeling'
        ]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        ],

    url=REPO_URL,
    author= 'Hiroshi Shinaoka',
    author_email='h.shinaoka@gmail.com',

    python_requires='>=3',
    install_requires=[
        'numpy',
        'scipy',
        'sparse_ir>=0.12.0',
        'admmsolver>=0.7.4',
    ],
    extras_require={
        'test': ['pytest', 'mypy'],
        },

    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    )
